/* 

                          Firewall Builder

                 Copyright (C) 2002 NetCitadel, LLC

  Author:  Vadim Kurland     vadim@vk.crocodile.org

  $Id: PolicyCompiler_ipt_optimizer.cc,v 1.11 2003/06/30 00:42:22 vkurland Exp $

  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "PolicyCompiler_ipt.hh"

#include "fwbuilder/FWObjectDatabase.hh"
#include "fwbuilder/RuleElement.hh"
#include "fwbuilder/IPService.hh"
#include "fwbuilder/ICMPService.hh"
#include "fwbuilder/TCPService.hh"
#include "fwbuilder/UDPService.hh"
#include "fwbuilder/Policy.hh"

#include "combinedAddress.hh"

#include <limits.h>

#include <iostream>
#include <iomanip>
#include <sstream>

#include <assert.h>

using namespace libfwbuilder;
using namespace fwcompiler;
using namespace std;


void PolicyCompiler_ipt::optimize1::optimizeForRuleElement(PolicyRule         *rule, 
                                                           const std::string  &re_type)
{
    RuleElement    *re=RuleElement::cast(rule->getFirstByType(re_type));
    int nre=re->size();

    PolicyRule     *r;

    string new_chain=PolicyCompiler_ipt::getNewTmpChainName(rule);

    r= PolicyRule::cast(compiler->dbcopy->create(PolicyRule::TYPENAME,true) );
    compiler->temp_ruleset->add(r);
    r->duplicate(rule);

    for (FWObject::iterator i=r->begin(); i!=r->end(); ++i) {
        if (RuleElement::cast(*i)!=NULL && (*i)->getTypeName()!=re_type) {
            RuleElement *nre=RuleElement::cast(*i);
            nre->clearChildren();  
            nre->setAnyElement();
        }
    }
    r->setStr("ipt_target",new_chain);
    tmp_queue.push_back(r);

    rule->setStr("ipt_chain",new_chain);
    re->clearChildren();  
    re->setAnyElement();
    tmp_queue.push_back(rule);
}

bool PolicyCompiler_ipt::optimize1::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementSrc *srcrel=rule->getSrc();
    RuleElementDst *dstrel=rule->getDst();
    RuleElementSrv *srvrel=rule->getSrv();

    int srcn=srcrel->size();
    int dstn=dstrel->size();
    int srvn=srvrel->size();

    if ( (srcn==1 && dstn==1) ||
         (dstn==1 && srvn==1) ||
         (srvn==1 && srcn==1) ) {
        tmp_queue.push_back(rule);
        return true;
    }

    if (srcn==1) srcn=INT_MAX;
    if (dstn==1) dstn=INT_MAX;
    if (srvn==1) srvn=INT_MAX;

    string re=RuleElementSrc::TYPENAME;

    if (srcn<=dstn && dstn<=srvn) {
        optimizeForRuleElement(rule,RuleElementSrc::TYPENAME);
        return true;
    }

    if (dstn<=srvn && srvn<=srcn) {
        optimizeForRuleElement(rule,RuleElementDst::TYPENAME);
        return true;
    }

    if (srvn<=srcn && srcn<=dstn) {
        optimizeForRuleElement(rule,RuleElementSrv::TYPENAME);
        return true;
    }


    tmp_queue.push_back(rule);

    return true;
}

bool PolicyCompiler_ipt::optimize2::processNext()
{
    PolicyCompiler_ipt *ipt_comp=dynamic_cast<PolicyCompiler_ipt*>(compiler);
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementSrv *srvrel=rule->getSrv();

    if (rule->getBool("final"))
    {
        if ( rule->getAction()==PolicyRule::Reject && ipt_comp->isActionOnRejectTCPRST(rule))
        {
// preserve service
            ;
        } else 
        {
            srvrel->clearChildren();
            srvrel->setAnyElement();
        }
    }

    tmp_queue.push_back(rule);

    return true;
}

/*
 *  this processor eliminates duplicate rules _generated for the same high level rule_
 *  This is different from processor PolicyCompiler_ipf::eliminateDuplicateRules, which
 *  finds and eliminates duplicate rules throughout the whole generated script.
 */
bool PolicyCompiler_ipt::optimize3::processNext()
{
    PolicyRule *rule;
    rule=getNext(); if (rule==NULL) return false;

    if (rule->isFallback() || rule->isHidden())
    {
        tmp_queue.push_back(rule);
        return true;
    }

    string lbl=rule->getLabel();

    std::deque<PolicyRule*>::iterator i;
    for (i=rules_seen_so_far[lbl].begin(); i!=rules_seen_so_far[lbl].end(); ++i)
    {
        PolicyRule *pr=PolicyRule::cast(*i);

        if ( pr->getInterfaceId()==rule->getInterfaceId() &&
             pr->getAction()==rule->getAction() &&
             pr->getDirection()==rule->getDirection() &&
             pr->getStr("ipt_target")==rule->getStr("ipt_target") &&
             pr->getStr("ipt_chain")==rule->getStr("ipt_chain") &&
             pr->getLabel()==rule->getLabel())
        {
            RuleElementSrc *srcrel1=rule->getSrc();
            RuleElementDst *dstrel1=rule->getDst();
            RuleElementSrv *srvrel1=rule->getSrv();

            RuleElementSrc *srcrel2=pr->getSrc();
            RuleElementDst *dstrel2=pr->getDst();
            RuleElementSrv *srvrel2=pr->getSrv();

            if (srcrel1->getNeg()!=srcrel2->getNeg() ||
                dstrel1->getNeg()!=dstrel2->getNeg() ||
                srvrel1->getNeg()!=srvrel2->getNeg() ) continue;

            Interval *ri1  = compiler->getFirstWhen(rule);
            Interval *ri2  = compiler->getFirstWhen(pr);

            if ( (ri1!=NULL && ri2==NULL) || (ri1==NULL && ri2!=NULL)) continue;
            if ( ri1!=NULL && ri2!=NULL && !( *ri1 == *ri2 ) ) continue;

            Address  *src1 = compiler->getFirstSrc(rule);  assert(src1);
            Address  *dst1 = compiler->getFirstDst(rule);  assert(dst1);
            Service  *srv1 = compiler->getFirstSrv(rule);  assert(srv1);

            Address  *src2 = compiler->getFirstSrc(pr);    assert(src2);
            Address  *dst2 = compiler->getFirstDst(pr);    assert(dst2);
            Service  *srv2 = compiler->getFirstSrv(pr);    assert(srv2);

            string physaddress1="";
            if (physAddress::isA(src1))      physaddress1= physAddress::cast(src1)->getPhysAddress();
            if (combinedAddress::isA(src1))  physaddress1= combinedAddress::cast(src1)->getPhysAddress();

            string physaddress2="";
            if (physAddress::isA(src2))      physaddress2= physAddress::cast(src2)->getPhysAddress();
            if (combinedAddress::isA(src2))  physaddress2= combinedAddress::cast(src2)->getPhysAddress();

            if ( physaddress1!=physaddress2) continue;

            if ( *src1 == *src2 && 
                 *dst1 == *dst2 && 
                 *srv1 == *srv2 ) return true;
        }
    }

    tmp_queue.push_back(rule);
    rules_seen_so_far[lbl].push_back(rule);

    return true;
}
